using System.Diagnostics;
using System.Security.Cryptography;
using System.Text;

class HOTP
{
    private static long GenerateHOTP(byte[] key, long counter, int digits = 6)
    {
        using var hmac = new HMACSHA512(key);
        byte[] counterBytes = BitConverter.GetBytes(counter);
        if (BitConverter.IsLittleEndian) Array.Reverse(counterBytes);

        byte[] hash = hmac.ComputeHash(counterBytes);

        int offset = hash[hash.Length - 1] & 0x0F;

        long nextEightBytesFromOffset = BitConverter.ToInt64(hash, offset) & 0x7FFFFFFFFFFFFFFFL;

        return nextEightBytesFromOffset % (long)Math.Pow(10, digits);
    }

    private static long CalculateCounter(long T0, long TX)
    {
        double T = (long)(DateTime.UtcNow - DateTime.UnixEpoch).TotalSeconds + T0;
        long CT = (long)Math.Floor((T - T0) / TX);
        return CT;
    }

    static void Main()
    {
        Stopwatch sp = new();
        sp.Start();

        int intervalInSeconds = 30;
        int T0 = 0;
        int digits = 8;
        
        for (int i = 0; i < 20; i++)
        {
            byte[] secretKey = Encoding.ASCII.GetBytes("i_am_secret_key");
            long counter = CalculateCounter(T0, intervalInSeconds);

            long hotpValue = GenerateHOTP(secretKey, counter, digits);

            Console.WriteLine($"elapsed {sp.ElapsedMilliseconds / 1000} seconds");
            Console.WriteLine($"HOTP value: {hotpValue:D8}");
            Thread.Sleep(10000);
        }
    }
}
